"""
Administration of Amazon S3 Resources through a boto3 client.
"""


def download_artifact(resource, bucket, key='alppbBuilder/alppb.zip',
                      local_path='alppb.zip'):
    """
    Fetches the artifact from Amazon S3 built by alppb .

    Parameters
    ----------
    resource : boto3.resources.factory.s3.ServiceResource
        A boto3 Resource interface for S3.
    bucket : str
        Name of the bucket the artifact exists in.
    key: str
        The S3 key of the artifact to download from the bucket. Defaults to
        "alppbBuilder/alppb.zip".
    local_path: str
        The local path where the artifact will be downloaded. Defaults to the
        current directory alppb is being run from.

    Returns
    -------
    """
    print("Downloading built package(s) to {}...".format(local_path))
    resource.meta.client.download_file(bucket, key, local_path)
    print("Download complete...")


def delete_artifact(client, bucket, key='alppbBuilder/alppb.zip'):
    """
    Deletes the artifact from Amazon S3 built by alppb.

    Parameters
    ----------
    client : botocore.client.S3
        A boto3 client for S3.
    bucket : str
        Name of the bucket the artifact exists in.
    key: str
        The S3 key of the artifact to delete from the bucket. Defaults to
        "alppbBuilder/alppb.zip".

    Returns
    -------
    """
    print("Deleting {} from S3 bucket {}...".format(key, bucket))
    client.delete_object(Bucket=bucket, Key=key)


def bucket_region(client, bucket):
    """
    Gets the region where this bucket exists.

    Parameters
    ----------
    client : botocore.client.S3
        A boto3 client for S3.

    Returns
    -------
    str
        Region of the bucket.
    """
    response = client.get_bucket_location(Bucket=bucket)
    location = response.get('LocationConstraint')
    if location is None:
        return 'us-east-1'
    return str(location)



"""
Administration of AWS CodeBuild Resources through a boto3 client.
"""
import time
import yaml


def determine_image(py_version):
    """
    Determines which Docker image to use for CodeBuild based on the Python
    version being used on AWS Lambda.

    Parameters
    ----------
    py_version : str
        Python version being used. Must be one of "2.7", "3.6", or "3.7"

    Returns
    -------
    str
        URI of the Docker image.
    """
    if py_version == "2.7":
        return "irlrobot/alppb-python27"
    if py_version == ("3.6" or "None"):
        return "irlrobot/alppb-python36"
    if py_version == "3.7":
        return "irlrobot/alppb-python37"

    # We should never get here, but fallback to Python 3.6.
    return "irlrobot/alppb-python36"


def pip_to_use(py_version):
    """
    Determines which pip command to use for CodeBuild based on the Python
    version being used on AWS Lambda.

    Parameters
    ----------
    py_version : str
        Python version being used. Must be one of "2.7", "3.6", or "3.7"

    Returns
    -------
    str
        Pip command that should be used.
    """
    if py_version == "2.7":
        return "pip-2.7"
    if py_version == ("3.6" or "None"):
        return "pip-3.6"
    if py_version == "3.7":
        return "pip3.7"

    # We should never get here, but fallback to pip-3.6.
    return "pip-3.6"


def generate_buildspec(package, py_version):
    """
    Creates a valid Buildspec, from a template, for an AWS CodeBuild project.
    The template requires a valid PyPi package to be specified.

    Parameters
    ----------
    package : str
        Name of the PyPi package to be built by AWS CodeBuild.

    py_version: str
        Python version being used. Must be one of "2.7", "3.6", or "3.7"

    Returns
    -------
    str
        The Buildspec in YAML.
    """
    return yaml.dump({
        "version": 0.2,
        "phases": {
            "build": {
                "commands": [
                    "{} install {} -t alppb".format(pip_to_use(py_version),
                                                    package),
                    "cd alppb/",
                    "zip -r ../alppb.zip *"
                ]
            }
        },
        "artifacts": {
            "files": [
                "alppb.zip"
            ]
        }
    })


def create_build_project(client, role, bucket, buildspec, image):
    """
    Creates a new AWS CodeBuild Project to build the PyPi package(s).

    Parameters
    ----------
    client : botocore.client.S3
        A boto3 client for S3.
    role : str
        Name of the IAM Role the AWS CodeBuild project should use.
    bucket : str
        Name of the bucket the build artifact will be put in.
    buildspec : str
        A valid AWS CodeBuild Buildspec in YAML. Use s3.generate_buildspec().
    image : str
        The Docker image to use for the CodeBuild project.

    Returns
    -------
    dict
        boto3 response object. See https://boto3.amazonaws.com/v1/documentation
        /api/latest/reference/services
        /codebuild.html#CodeBuild.Client.create_project for more information.
    """
    print("Creating CodeBuild project...")
    try:
        response = client.create_project(
            name='alppbBuilder',
            source={
                'type': 'NO_SOURCE',
                'buildspec': buildspec,
            },
            artifacts={
                'type': 'S3',
                'location': bucket,
            },
            environment={
                'type': 'LINUX_CONTAINER',
                'image': image,
                'computeType': 'BUILD_GENERAL1_SMALL',
            },
            serviceRole=role,
        )
    except client.exceptions.ResourceAlreadyExistsException:
        print(">>alppbBuilder project already exists, overwriting...")
        response = client.update_project(
            name='alppbBuilder',
            source={
                'type': 'NO_SOURCE',
                'buildspec': buildspec,
            },
            artifacts={
                'type': 'S3',
                'location': bucket,
            },
            environment={
                'type': 'LINUX_CONTAINER',
                'image': image,
                'computeType': 'BUILD_GENERAL1_SMALL',
            },
            serviceRole=role,
        )

    return response


def delete_build_project(client):
    """
    Deletes the alppb AWS CodeBuild project.

    Parameters
    ----------
    client : botocore.client.codebuild
        A boto3 client for CodeBuild.

    Returns
    -------
    """
    print("Deleting CodeBuild project...")
    client.delete_project(name="alppbBuilder")


def wait_for_build_to_complete(client, build_id):
    """
    Keeps checking until a build completes.

    Parameters
    ----------
    client : botocore.client.codebuild
        A boto3 client for CodeBuild.

    build_id : str
        The ID of the AWS CodeBuild job to poll.

    Returns
    -------
    """
    response = client.batch_get_builds(ids=[build_id])
    # batch_get_builds() will return an array with one element.
    status = str(response.get('builds')[0].get('buildStatus'))
    if status == 'SUCCEEDED':
        print("Build completed...")
        return
    # TODO add case when status is FAILED
    elif status == 'IN_PROGRESS':
        print(">>Build {}, waiting 10 seconds...".format(status))
        time.sleep(10)
        # Recursively call until the build is done.
        return wait_for_build_to_complete(client, build_id)
    else:
        print("ERROR: Build did not complete, status is: {}. Check "
              "the AWS console for more information on {}".format(
                  status, build_id
              ))
        exit(1)


def build_artifact(client):
    """
    Start a build and wait until it's done.

    Parameters
    ----------
    client : botocore.client.codebuild
        A boto3 client for CodeBuild.

    Returns
    -------
    """
    print("Submitting a build job for the specified package(s)...")
    response = client.start_build(projectName='alppbBuilder')
    build_id = str(response.get('build').get('id'))
    print(">>Build ID is {}".format(build_id))
    wait_for_build_to_complete(client, build_id)



#!/usr/bin/env python
"""
alppb
Builds Python modules on Amazon Linux using AWS CodeBuild and downloads them.
"""
__author__ = "Josh Campbell"
__version__ = "0.2.0"
__license__ = "MIT"
import argparse
import boto3
from botocore.exceptions import NoRegionError
from . import codebuild
from . import iam
from . import s3


def check_for_boto_credentials():
    """
    Checks if credentials for boto3 to use are available.

    Parameters
    ----------

    Returns
    -------
    """
    session = boto3.Session()
    credentials = session.get_credentials()
    if credentials is None:
        print("ERROR: boto3 credentials missing. "
              "Does ~/.aws/credentials exist?")
        exit(1)


def create_client(service, region):
    """
    Creates a boto3 client object for the specified service.

    Parameters
    ----------
    service : str
        The name of the AWS service to create the client for.

    Returns
    -------
    botocore.client.service
        See https://boto3.amazonaws.com/v1/documentation/api/latest/guide
        /clients.html for more information.
    """
    print("Creating boto3 client for {}...".format(service))
    try:
        return boto3.client(service, region_name=region)
    except NoRegionError:
        print("ERROR: boto3 cannot find a default profile to use. Try running "
              "`aws configure`. To see how boto3 loads configuration, see: "
              "https://boto3.amazonaws.com/v1/documentation/api/latest/guide/"
              "configuration.html#configuring-credentials")
        exit(1)


def create_resource(service, region):
    """
    Creates a boto3 resource object for the specified service.

    Parameters
    ----------
    service : str
        The name of the AWS service to create the resource for.

    Returns
    -------
    boto3.resources.factory.service.ServiceResource
        See https://boto3.amazonaws.com/v1/documentation/api/latest/guide
        /resources.html for more information.
    """
    print("Creating boto3 resource for {}...".format(service))
    try:
        return boto3.resource(service, region_name=region)
    except NoRegionError:
        print("ERROR: boto3 cannot find a default profile to use. Try running "
              "`aws configure`. To see how boto3 loads configuration, see: "
              "https://boto3.amazonaws.com/v1/documentation/api/latest/guide/"
              "configuration.html#configuring-credentials")
        exit(1)


def parse_args():
    """ Setup ArgumentParser """
    parser = argparse.ArgumentParser()
    parser.add_argument("package", help="The PyPi package you want to build.",
                        type=str)
    parser.add_argument("bucket", help="Name of the S3 bucket to use.",
                        type=str)
    parser.add_argument("--region", help="The AWS region to use.", type=str,
                        choices=boto3.session.Session().get_available_regions(
                            'codebuild'))
    parser.add_argument("--python",
                        help="The Python version to use. Defaults to 3.6.",
                        type=str, choices=["2.7", "3.6", "3.7"])

    return parser.parse_args()


def main():
    """ Main entry point of the app """
    print("Starting alppb...")
    check_for_boto_credentials()

    # Parse args and get values used in functions below.
    args = parse_args()
    package = args.package
    bucket = args.bucket
    # If region is None boto3 will determine the region to use. See more at,
    # https://boto3.amazonaws.com/v1/documentation/api/latest/guide
    # /configuration.html#configuring-credentials
    region = args.region
    py_version = args.python

    # Create boto3 clients/resources used below.
    iam_client = create_client('iam', region)
    codebuild_client = create_client('codebuild', region)
    s3_resource = create_resource('s3', region)
    s3_client = create_client('s3', region)

    # Check if S3 bucket is in the same region as target CodeBuild region.
    bucket_region = s3.bucket_region(s3_client, bucket)
    codebuild_region = codebuild_client._client_config.__dict__\
        .get('_user_provided_options').get('region_name')
    if bucket_region != codebuild_region:
        print("ERROR: Bucket and CodeBuild project must be in the same "
              "region. Bucket is in {}, but the region being used for "
              "CodeBuild is {}. Recommended Action: Set the --region flag "
              "to {}. e.g. `alppb {} {} --region {}`".format(
                bucket_region, codebuild_region, bucket_region, package,
                bucket, bucket_region))
        exit(1)

    # Create alppb resources.
    role = iam.create_role(iam_client, bucket)
    buildspec = codebuild.generate_buildspec(package, py_version)
    codebuild.create_build_project(codebuild_client, role, bucket, buildspec,
                                   codebuild.determine_image(py_version))
    codebuild.build_artifact(codebuild_client)

    # Download the artifact.
    s3.download_artifact(s3_resource, bucket)

    # Cleanup phase.
    codebuild.delete_build_project(codebuild_client)
    iam.delete_role(iam_client)
    s3.delete_artifact(s3_client, bucket)

    print("SUCCESS")
    exit(0)


if __name__ == "__main__":
    """ This is executed when run from the command line """
    main()



import sys
from unittest.mock import patch
import pytest
from alppb.alppb import main


def test_creating_default_package_with_valid_input():
    testargs = ["alppb", "mlbgame", "rebukethe.net"]
    with patch.object(sys, 'argv', testargs):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            main()

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0


def test_creating_default_package_with_bad_bucket():
    testargs = ["alppb", "987dfdsf#&^%#*", "$^&$&*sfdsfsdf"]
    with patch.object(sys, 'argv', testargs):
        with pytest.raises(Exception):
            main()


def test_creating_python27_package_with_valid_input():
    testargs = ["alppb", "mlbgame", "rebukethe.net", "--python", "2.7"]
    with patch.object(sys, 'argv', testargs):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            main()

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0


def test_creating_python27_package_with_bad_bucket():
    testargs = ["alppb", "987dfdsf#&^%#*", "$^&$&*sfdsfsdf", "--python", "2.7"]
    with patch.object(sys, 'argv', testargs):
        with pytest.raises(Exception):
            main()


def test_creating_python36_package_with_valid_input():
    testargs = ["alppb", "mlbgame", "rebukethe.net", "--python", "3.6"]
    with patch.object(sys, 'argv', testargs):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            main()

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0


def test_creating_python36_package_with_bad_bucket():
    testargs = ["alppb", "987dfdsf#&^%#*", "$^&$&*sfdsfsdf", "--python", "3.6"]
    with patch.object(sys, 'argv', testargs):
        with pytest.raises(Exception):
            main()


def test_creating_python37_package_with_valid_input():
    testargs = ["alppb", "mlbgame", "rebukethe.net", "--python", "3.7"]
    with patch.object(sys, 'argv', testargs):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            main()

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0


def test_creating_python37_package_with_bad_bucket():
    testargs = ["alppb", "987dfdsf#&^%#*", "$^&$&*sfdsfsdf", "--python", "3.7"]
    with patch.object(sys, 'argv', testargs):
        with pytest.raises(Exception):
            main()



"""
Administration of AWS IAM Resources through a boto3 client.
"""
import json
import time


def create_role(client, bucket):
    """
    Creates an IAM Role to use with AWS CodeBuild

    Parameters
    ----------
    client : botocore.client.iam
        A boto3 client for IAM.

    bucket : str
        Name of an existing S3 bucket.

    Returns
    -------
    str
        ARN of the newly created IAM Role.
    """
    print("Creating IAM Role...")
    try:
        response = client.create_role(
            RoleName='alppbBuilderRole',
            AssumeRolePolicyDocument=json.dumps({
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {
                            "Service": "codebuild.amazonaws.com"
                        },
                        "Action": "sts:AssumeRole"
                    }
                ]
            }),
            Description="This role is used by "
                        "https://github.com/irlrobot/alppb",
        )
    except client.exceptions.EntityAlreadyExistsException:
        print(">>alppbBuilderRole already exists, skipping...")
        response = client.get_role(RoleName='alppbBuilderRole')

    add_role_policy(client, bucket)

    # TODO be smarter about checking if the role is ready
    print(">>Waiting 10 seconds for policy propagation before continuing...")
    time.sleep(10)

    return str(response.get('Role').get('Arn'))


def generate_role_policy(bucket):
    """
    Generates a valid IAM Role policy from a template. The template requires
    the name of an existing S3 bucket.

    Parameters
    ----------
    bucket : str
        Name of an existing S3 bucket.

    Returns
    -------
    str
        The IAM Policy document in JSON.
    """
    return json.dumps({
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "CloudWatchLogsPolicy",
                "Effect": "Allow",
                "Action": [
                    "logs:CreateLogGroup",
                    "logs:CreateLogStream",
                    "logs:PutLogEvents"
                ],
                "Resource": [
                    "*"
                ]
            },
            {
                "Effect": "Allow",
                "Action": [
                    "s3:ListBucket"
                ],
                "Resource": [
                    "arn:aws:s3:::{}".format(bucket)
                ]
            },
            {
                "Effect": "Allow",
                "Action": [
                    "s3:PutObject",
                    "s3:GetObject",
                    "s3:DeleteObject"
                ],
                "Resource": [
                    "arn:aws:s3:::{}/*".format(bucket)
                ]
            }
        ]
    })


def add_role_policy(client, bucket):
    """
    Adds an IAM Policy to a newly created IAM Role from iam.create_role().

    Parameters
    ----------
    client : botocore.client.iam
        A boto3 client for IAM.

    bucket : str
        Name of an existing S3 bucket.

    Returns
    -------
    """
    print(">>Attaching Policy to the the IAM Role...")
    client.put_role_policy(
        RoleName='alppbBuilderRole',
        PolicyName='alppbBuilderPolicy',
        PolicyDocument=generate_role_policy(bucket),
    )


def delete_role(client):
    """
    Deletes an IAM Role created from iam.create_role().

    Parameters
    ----------
    client : botocore.client.iam
        A boto3 client for IAM.

    Returns
    -------
    """
    print("Deleting IAM Role...")
    # AWS API wants all role policies deleted before the role itself.
    # TODO get all policies on the role and then delete in case it was modified
    client.delete_role_policy(
        RoleName='alppbBuilderRole',
        PolicyName='alppbBuilderPolicy'
    )
    client.delete_role(RoleName='alppbBuilderRole')
